#! /usr/bin/env python3

import torch

from .generic_pair_loss import GenericPairLoss
from ..utils import loss_and_miner_utils as lmu, common_functions as c_f


class LiftedStructureLoss(GenericPairLoss):
    def __init__(self, neg_margin, pos_margin=0, **kwargs):
        super().__init__(use_similarity=False, mat_based_loss=False, **kwargs)
        self.neg_margin = neg_margin
        self.pos_margin = pos_margin

    def _compute_loss(self, pos_pairs, neg_pairs, indices_tuple):
        a1, p, a2, _ = indices_tuple

        if len(a1) > 0 and len(a2) > 0:
            pos_pairs = pos_pairs.unsqueeze(1)
            n_per_p = ((a2.unsqueeze(0) == a1.unsqueeze(1)) | (a2.unsqueeze(0) == p.unsqueeze(1))).float()
            neg_pairs = neg_pairs*n_per_p
            keep_mask = (~(n_per_p==0)).float()

            neg_pairs_loss = lmu.logsumexp(self.neg_margin-neg_pairs, keep_mask=keep_mask, add_one=False, dim=1)
            loss_per_pos_pair = neg_pairs_loss + (pos_pairs - self.pos_margin)
            loss_per_pos_pair = torch.relu(loss_per_pos_pair)**2
            loss_per_pos_pair /= 2 # divide by 2 since each positive pair will be counted twice
            return {"loss": {"losses": loss_per_pos_pair, "indices": (a1, p), "reduction_type": "pos_pair"}}
        return self.zero_losses()



class GeneralizedLiftedStructureLoss(GenericPairLoss):
    # The 'generalized' lifted structure loss shown on page 4
    # of the "in defense of triplet loss" paper
    # https://arxiv.org/pdf/1703.07737.pdf
    def __init__(self, neg_margin, pos_margin=0, **kwargs):
        super().__init__(use_similarity=False, mat_based_loss=True, **kwargs)
        self.neg_margin = neg_margin
        self.pos_margin = pos_margin

    def _compute_loss(self, mat, pos_mask, neg_mask):
        pos_loss = lmu.logsumexp(mat - self.pos_margin, keep_mask=pos_mask, add_one=False)
        neg_loss = lmu.logsumexp(self.neg_margin - mat, keep_mask=neg_mask, add_one=False)
        return {"loss": {"losses": torch.relu(pos_loss+neg_loss), "indices": c_f.torch_arange_from_size(mat), "reduction_type": "element"}}



#! /usr/bin/env python3

import torch
from ..utils import common_functions as c_f
from ..utils.module_with_records_and_reducer import ModuleWithRecordsAndReducer

class BaseWeightRegularizer(ModuleWithRecordsAndReducer):
    def __init__(self, normalize_weights=True, **kwargs):
        super().__init__(**kwargs)
        self.normalize_weights = normalize_weights
        self.add_to_recordable_attributes(name="avg_weight_norm", is_stat=True)

    def compute_loss(self, weights):
        raise NotImplementedError

    def forward(self, weights):
        """
        weights should have shape (num_classes, embedding_size)
        """
        self.reset_stats()
        if self.normalize_weights:
            weights = torch.nn.functional.normalize(weights, p=2, dim=1)
        self.weight_norms = torch.norm(weights, p=2, dim=1)
        self.avg_weight_norm = torch.mean(self.weight_norms)
        loss_dict = self.compute_loss(weights)
        return self.reducer(loss_dict, weights, c_f.torch_arange_from_size(weights))


#! /usr/bin/env python3

from .base_miner import BaseTupleMiner
from ..utils import loss_and_miner_utils as lmu
import torch


class TripletMarginMiner(BaseTupleMiner):
    """
    Returns triplets that violate the margin
    Args:
    	margin
    	type_of_triplets: options are "all", "hard", or "semihard".
    		"all" means all triplets that violate the margin
    		"hard" is a subset of "all", but the negative is closer to the anchor than the positive
    		"semihard" is a subset of "all", but the negative is further from the anchor than the positive
            "easy" is all triplets that are not in "all"
    """
    def __init__(self, margin, type_of_triplets="all", **kwargs):
        super().__init__(**kwargs)
        self.margin = margin
        self.add_to_recordable_attributes(list_of_names=["avg_triplet_margin", "pos_pair_dist", "neg_pair_dist"], is_stat=True)
        self.type_of_triplets = type_of_triplets

    def mine(self, embeddings, labels, ref_emb, ref_labels):
        anchor_idx, positive_idx, negative_idx  = lmu.get_all_triplets_indices(labels, ref_labels)
        anchors, positives, negatives = embeddings[anchor_idx], ref_emb[positive_idx], ref_emb[negative_idx]
        ap_dist = torch.nn.functional.pairwise_distance(anchors, positives, 2)
        an_dist = torch.nn.functional.pairwise_distance(anchors, negatives, 2)
        triplet_margin = an_dist - ap_dist
        self.pos_pair_dist = torch.mean(ap_dist).item()
        self.neg_pair_dist = torch.mean(an_dist).item()
        self.avg_triplet_margin = torch.mean(triplet_margin).item()
        if self.type_of_triplets == "easy":
            threshold_condition = triplet_margin > self.margin
        else:
            threshold_condition = triplet_margin <= self.margin
            if self.type_of_triplets == "hard":
                threshold_condition &= an_dist <= ap_dist
            elif self.type_of_triplets == "semihard":
                threshold_condition &= an_dist > ap_dist
        return anchor_idx[threshold_condition], positive_idx[threshold_condition], negative_idx[threshold_condition]


from ..reducers import MeanReducer, MultipleReducers
from .module_with_records import ModuleWithRecords

class ModuleWithRecordsAndReducer(ModuleWithRecords):
    def __init__(self, reducer=None, **kwargs):
        super().__init__(**kwargs)
        self.reducer = self.get_reducer() if reducer is None else reducer

    def get_default_reducer(self):
        return MeanReducer()

    def get_reducer(self):
        reducer = self.get_default_reducer()
        if isinstance(reducer, MultipleReducers) or len(self.sub_loss_names()) == 1:
            return reducer
        return MultipleReducers({k:self.get_default_reducer() for k in self.sub_loss_names()})

    def sub_loss_names(self):
        return ["loss"]


import torch
from ..utils import loss_and_miner_utils as lmu, common_functions as c_f
from .base_metric_loss_function import BaseMetricLossFunction
from ..reducers import AvgNonZeroReducer, MultipleReducers, MeanReducer


def SNR_dist(x, y, dim):
    return torch.var(x-y, dim=dim) / torch.var(x, dim=dim)


class SignalToNoiseRatioContrastiveLoss(BaseMetricLossFunction):

    def __init__(self, pos_margin, neg_margin, regularizer_weight, **kwargs):
        super().__init__(**kwargs)
        self.pos_margin = pos_margin
        self.neg_margin = neg_margin
        self.regularizer_weight = regularizer_weight
        self.add_to_recordable_attributes(name="feature_distance_from_zero_mean_distribution", is_stat=True)
        
    def compute_loss(self, embeddings, labels, indices_tuple):
        a1, p, a2, n = lmu.convert_to_pairs(indices_tuple, labels)
        pos_loss, neg_loss, reg_loss = 0, 0, 0
        if len(a1) > 0:
            pos_loss = self.get_per_pair_loss(embeddings[a1], embeddings[p], self.pos_margin, 1)
        if len(a2) > 0:
            neg_loss = self.get_per_pair_loss(embeddings[a2], embeddings[n], self.neg_margin, -1)
        self.feature_distance_from_zero_mean_distribution = torch.abs(torch.sum(embeddings, dim=1))
        if self.regularizer_weight > 0:
            reg_loss = self.regularizer_weight * self.feature_distance_from_zero_mean_distribution
        
        loss_dict = {"pos_loss": {"losses": pos_loss, "indices": (a1, p), "reduction_type": "pos_pair"}, 
                    "neg_loss": {"losses": neg_loss, "indices": (a2, n), "reduction_type": "neg_pair"},
                    "reg_loss": {"losses": reg_loss, "indices": c_f.torch_arange_from_size(embeddings), "reduction_type": "element"}}

        return loss_dict


    def get_per_pair_loss(self, anchors, others, margin, before_relu_multiplier):
        d = SNR_dist(anchors, others, dim=1)
        return torch.nn.functional.relu((d-margin)*before_relu_multiplier)

    def get_default_reducer(self):
        return MultipleReducers({"reg_loss": MeanReducer()}, default_reducer=AvgNonZeroReducer())

    def sub_loss_names(self):
        return ["pos_loss", "neg_loss", "reg_loss"]


#! /usr/bin/env python3

from .weight_regularizer_mixin import WeightRegularizerMixin
from .base_metric_loss_function import BaseMetricLossFunction
from ..utils import loss_and_miner_utils as lmu, common_functions as c_f
import scipy.special
import torch
import math
import numpy as np

class LargeMarginSoftmaxLoss(WeightRegularizerMixin, BaseMetricLossFunction):
    """
    Implementation of https://arxiv.org/pdf/1612.02295.pdf
    """
    def __init__(self, margin, num_classes, embedding_size, scale=1, normalize_weights=False, scale_logits_by_magnitudes=True, **kwargs):
        super().__init__(**kwargs)
        self.margin = margin
        self.num_classes = num_classes
        self.scale = scale
        self.normalize_weights = normalize_weights
        self.scale_logits_by_magnitudes = scale_logits_by_magnitudes
        self.add_to_recordable_attributes(name="avg_angle", is_stat=True)
        self.init_margin()
        self.W = torch.nn.Parameter(torch.randn(embedding_size, num_classes))
        self.cross_entropy = torch.nn.CrossEntropyLoss(reduction='none')

    def init_margin(self):
        self.margin = int(self.margin)
        self.max_n = (self.margin // 2)
        ## For the trigonometric multiple-angle formula ##
        self.n_range = torch.FloatTensor([n for n in range(0, self.max_n+1)])
        self.margin_choose_n = torch.FloatTensor([scipy.special.binom(self.margin, 2*n) for n in self.n_range])
        self.cos_powers = torch.FloatTensor([self.margin-(2*n) for n in self.n_range])
        self.alternating = torch.FloatTensor([(-1)**n for n in self.n_range])

    def get_cos_with_margin(self, cosine):
        cosine = cosine.unsqueeze(1)
        for attr in ["n_range", "margin_choose_n", "cos_powers", "alternating"]:
            setattr(self, attr, getattr(self, attr).to(cosine.device))
        cos_powered = cosine**self.cos_powers
        sin_powered = (1-cosine**2)**self.n_range
        terms = self.alternating*self.margin_choose_n*cos_powered*sin_powered # Equation 7 in the paper
        return torch.sum(terms, dim=1)

    def get_weights(self):
        if self.normalize_weights:
            return torch.nn.functional.normalize(self.W, p=2, dim=0)
        return self.W

    def get_cosine(self, embeddings):
        weights = self.get_weights()
        self.weight_norms = torch.norm(weights, p=2, dim=0)
        # self.embedding_norms is computed in BaseMetricLossFunction
        self.product_of_magnitudes = self.weight_norms.unsqueeze(0)*self.embedding_norms.unsqueeze(1)
        return torch.matmul(embeddings, weights) / self.product_of_magnitudes

    def get_angles(self, cosine_of_target_classes):
        angles = torch.acos(torch.clamp(cosine_of_target_classes, -1 + 1e-7, 1 - 1e-7))
        self.avg_angle = np.degrees(torch.mean(angles).item())
        return angles

    def get_target_mask(self, embeddings, labels):
        batch_size = labels.size(0)
        mask = torch.zeros(batch_size, self.num_classes).to(embeddings.device)
        mask[torch.arange(batch_size), labels] = 1
        return mask

    def modify_cosine_of_target_classes(self, cosine_of_target_classes, *args):
        _, _, labels, _ = args
        cos_with_margin = self.get_cos_with_margin(cosine_of_target_classes)
        angles = self.get_angles(cosine_of_target_classes)
        with torch.no_grad():
            k = (angles / (math.pi / self.margin)).floor() # Equation 6: angles needs to be between [k*pi/m and (k+1)*pi/m]
        return ((-1)**k)*cos_with_margin - (2*k)

    def compute_loss(self, embeddings, labels, indices_tuple):
        miner_weights = lmu.convert_to_weights(indices_tuple, labels)
        mask = self.get_target_mask(embeddings, labels)
        cosine = self.get_cosine(embeddings)
        cosine_of_target_classes = cosine[mask == 1]
        modified_cosine_of_target_classes = self.modify_cosine_of_target_classes(cosine_of_target_classes, cosine, embeddings, labels, mask)
        diff = (modified_cosine_of_target_classes - cosine_of_target_classes).unsqueeze(1)
        logits = cosine + (mask*diff)
        if self.scale_logits_by_magnitudes:
            logits = logits * self.product_of_magnitudes
        unweighted_loss = self.cross_entropy(logits * self.scale, labels)
        miner_weighted_loss = unweighted_loss*miner_weights
        loss_dict = {"loss": {"losses": miner_weighted_loss, "indices": c_f.torch_arange_from_size(embeddings), "reduction_type": "element"}}
        loss_dict["reg_loss"] = self.regularization_loss(self.W.t())
        return loss_dict


    def sub_loss_names(self):
        return ["loss", "reg_loss"]


#! /usr/bin/env python3

import torch
from ..utils import common_functions as c_f
from ..utils.module_with_records_and_reducer import ModuleWithRecordsAndReducer

class BaseMetricLossFunction(ModuleWithRecordsAndReducer):
    def __init__(self, normalize_embeddings=True, **kwargs):
        super().__init__(**kwargs)
        self.normalize_embeddings = normalize_embeddings
        self.add_to_recordable_attributes(name="avg_embedding_norm", is_stat=True)

    def compute_loss(self, embeddings, labels, indices_tuple=None):
        """
        This has to be implemented and is what actually computes the loss.
        """
        raise NotImplementedError

    def forward(self, embeddings, labels, indices_tuple=None):
        """
        Args:
            embeddings: tensor of size (batch_size, embedding_size)
            labels: tensor of size (batch_size)
            indices_tuple: tuple of size 3 for triplets (anchors, positives, negatives)
                            or size 4 for pairs (anchor1, postives, anchor2, negatives)
                            Can also be left as None
        Returns: the loss (float)
        """
        self.reset_stats()
        c_f.assert_embeddings_and_labels_are_same_size(embeddings, labels)
        labels = labels.to(embeddings.device)
        if self.normalize_embeddings:
            embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1)
        self.embedding_norms = torch.norm(embeddings, p=2, dim=1)
        self.avg_embedding_norm = torch.mean(self.embedding_norms)

        loss_dict = self.compute_loss(embeddings, labels, indices_tuple)
        return self.reducer(loss_dict, embeddings, labels)

    def zero_loss(self):
        return {"losses": 0, "indices": None, "reduction_type": "already_reduced"}

    def zero_losses(self):
        return {loss_name: self.zero_loss() for loss_name in self.sub_loss_names()}


class MultipleLosses(torch.nn.Module):
    def __init__(self, losses, weights=None):
        super().__init__()
        self.is_dict = isinstance(losses, dict)
        self.losses = torch.nn.ModuleDict(losses) if self.is_dict else torch.nn.ModuleList(losses)
        if weights is not None:
            assert isinstance(weights, dict) if self.is_dict else c_f.is_list_or_tuple(weights)
            self.weights = weights
        else:
            self.weights = {k:1 for k in self.losses.keys()} if self.is_dict else [1]*len(losses)


    def forward(self, embeddings, labels, indices_tuple=None):
        total_loss = 0
        iterable = self.losses.items() if self.is_dict else enumerate(self.losses)
        if self.is_dict:
            for i, loss_func in iterable:
                total_loss += loss_func(embeddings, labels, indices_tuple)*self.weights[i]
        return total_loss


from .large_margin_softmax_loss import LargeMarginSoftmaxLoss

class SphereFaceLoss(LargeMarginSoftmaxLoss):
	# implementation of https://arxiv.org/pdf/1704.08063.pdf
    def __init__(self, normalize_weights=True, **kwargs):
        super().__init__(normalize_weights=normalize_weights, **kwargs)
        assert self.normalize_weights, "SphereFace requires that the weights are normalized"


#! /usr/bin/env python3

from .base_metric_loss_function import BaseMetricLossFunction
import torch
from ..utils import loss_and_miner_utils as lmu, common_functions as c_f
from ..reducers import DivisorReducer

class MarginLoss(BaseMetricLossFunction):

    def __init__(self, margin, nu, beta, triplets_per_anchor="all", learn_beta=False, num_classes=None, **kwargs):
        super().__init__(**kwargs)
        self.margin = margin
        self.nu = nu
        self.learn_beta = learn_beta
        self.initialize_beta(beta, num_classes)
        self.triplets_per_anchor = triplets_per_anchor
        self.add_to_recordable_attributes(list_of_names=["beta"])
        
    def compute_loss(self, embeddings, labels, indices_tuple):
        indices_tuple = lmu.convert_to_triplets(indices_tuple, labels, self.triplets_per_anchor)
        anchor_idx, positive_idx, negative_idx = indices_tuple
        if len(anchor_idx) == 0:
            return self.zero_losses()
        anchors, positives, negatives = embeddings[anchor_idx], embeddings[positive_idx], embeddings[negative_idx]

        beta = self.beta if len(self.beta) == 1 else self.beta[labels[anchor_idx]]
        beta = beta.to(embeddings.device)
        
        d_ap = torch.nn.functional.pairwise_distance(positives, anchors, p=2)
        d_an = torch.nn.functional.pairwise_distance(negatives, anchors, p=2)

        pos_loss = torch.nn.functional.relu(d_ap - beta + self.margin)
        neg_loss = torch.nn.functional.relu(beta - d_an + self.margin)

        num_pos_pairs = (pos_loss > 0.0).nonzero().size(0)
        num_neg_pairs = (neg_loss > 0.0).nonzero().size(0)

        divisor_summands = {"num_pos_pairs": num_pos_pairs, "num_neg_pairs": num_neg_pairs}

        margin_loss = pos_loss + neg_loss

        loss_dict = {"margin_loss": {"losses": margin_loss, "indices": indices_tuple, "reduction_type": "triplet", "divisor_summands": divisor_summands}, 
                    "beta_reg_loss": self.compute_reg_loss(beta, anchor_idx, divisor_summands)}

        return loss_dict

    def compute_reg_loss(self, beta, anchor_idx, divisor_summands):
        if self.learn_beta:
            loss = beta * self.nu
            if len(self.beta) == 1:
                return {"losses": loss, "indices": None, "reduction_type": "already_reduced"}
            else:
                return {"losses": loss, "indices": anchor_idx, "reduction_type": "element", "divisor_summands": divisor_summands}
        return self.zero_loss()

    def sub_loss_names(self):
        return ["margin_loss", "beta_reg_loss"]

    def get_default_reducer(self):
        return DivisorReducer()

    def initialize_beta(self, beta, num_classes):
        self.beta = torch.tensor([float(beta)])
        if num_classes:
            self.beta = torch.ones(num_classes) * self.beta
        if self.learn_beta:
            self.beta = torch.nn.Parameter(self.beta)


class WeightRegularizerMixin:
    def __init__(self, regularizer=None, reg_weight=1, **kwargs):
        super().__init__(**kwargs)
        self.regularizer = regularizer
        self.reg_weight = reg_weight

    def regularization_loss(self, weights):
        if self.regularizer is None:
            loss = 0
        else:
            loss = self.regularizer(weights) * self.reg_weight
        return {"losses": loss, "indices": None, "reduction_type": "already_reduced"}


from .base_reducer import BaseReducer
import torch


class ThresholdReducer(BaseReducer):
    def __init__(self, threshold, **kwargs):
        super().__init__(**kwargs)
        self.threshold = threshold
        self.add_to_recordable_attributes(name="threshold")

    def element_reduction(self, losses, *_):
        return self.element_reduction_helper(losses, "elements_above_threshold")
    
    def pos_pair_reduction(self, losses, *args):
        return self.element_reduction_helper(losses, "pos_pairs_above_threshold")

    def neg_pair_reduction(self, losses, *args):
        return self.element_reduction_helper(losses, "neg_pairs_above_threshold")

    def triplet_reduction(self, losses, *args):
        return self.element_reduction_helper(losses, "triplets_above_threshold")

    def element_reduction_helper(self, losses, attr_name):
        threshold_condition = losses > self.threshold
        num_above_threshold = torch.sum(threshold_condition)
        if num_above_threshold >= 1:
            loss = torch.mean(losses[threshold_condition])
        else:
            loss = torch.mean(losses)*0 # set loss to 0
        self.add_to_recordable_attributes(name=attr_name, is_stat=True)
        setattr(self, attr_name, num_above_threshold)
        return loss


#! /usr/bin/env python3

from .generic_pair_loss import GenericPairLoss
import torch 
from ..utils import loss_and_miner_utils as lmu, common_functions as c_f
from ..reducers import AvgNonZeroReducer


class CircleLoss(GenericPairLoss):
    """
    Circle loss for pairwise labels only. Support for class-level labels will be added 
    in the future.
    
    Args:
    m:  The relaxation factor that controls the radious of the decision boundary.
    gamma: The scale factor that determines the largest scale of each similarity score.

    According to the paper, the suggested default values of m and gamma are:

    Face Recognition: m = 0.25, gamma = 256
    Person Reidentification: m = 0.25, gamma = 256
    Fine-grained Image Retrieval: m = 0.4, gamma = 80

    By default, we set m = 0.4 and gamma = 80
    """
    def __init__(
        self, 
        m=0.4,
        gamma=80,
        **kwargs
    ):
        super().__init__(use_similarity=True, mat_based_loss=True, **kwargs)
        self.m = m 
        self.gamma = gamma 
        self.soft_plus = torch.nn.Softplus(beta=1)
        self.op = 1 + self.m 
        self.on = -self.m
        self.delta_p = 1-self.m 
        self.delta_n = self.m 

    def _compute_loss(self, mat, pos_mask, neg_mask):
        pos_mask_bool = pos_mask.bool()
        neg_mask_bool = neg_mask.bool()
        anchor_positive = mat[pos_mask_bool]
        anchor_negative = mat[neg_mask_bool]
        new_mat = torch.zeros_like(mat)

        new_mat[pos_mask_bool] = -self.gamma * torch.relu(self.op - anchor_positive.detach()) * (anchor_positive - self.delta_p)
        new_mat[neg_mask_bool] = self.gamma * torch.relu(anchor_negative.detach() - self.on) * (anchor_negative - self.delta_n)

        losses = self.soft_plus(lmu.logsumexp(new_mat, keep_mask=pos_mask, add_one=False, dim=1) + lmu.logsumexp(new_mat, keep_mask=neg_mask, add_one=False, dim=1))

        zero_rows = torch.where((torch.sum(pos_mask, dim=1)==0) | (torch.sum(neg_mask, dim=1) == 0))[0]
        final_mask = torch.ones_like(losses)
        final_mask[zero_rows] = 0
        losses = losses*final_mask
        return {"loss": {"losses": losses, "indices": c_f.torch_arange_from_size(new_mat), "reduction_type": "element"}}

    def get_default_reducer(self):
        return AvgNonZeroReducer()


import torch
from . import common_functions as c_f


class ModuleWithRecords(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def add_to_recordable_attributes(self, name=None, list_of_names=None, is_stat=False):
        c_f.add_to_recordable_attributes(self, name=name, list_of_names=list_of_names, is_stat=is_stat)

    def reset_stats(self):
        c_f.reset_stats(self)
